from brownie import chain
from ypricemagic.networks import Network
from ypricemagic.price_modules.balancer.v2.pool import BalancerV2Pool
from ypricemagic.price_modules.balancer.v2.vault import BalancerV2Vault

BALANCER_V2_VAULTS = {
    Network.Mainnet: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Fantom: [
        '0x20dd72Ed959b6147912C2e529F0a0C651c33c9ce',
    ]
}

BALANCER_V2_VAULTS = [BalancerV2Vault(vault) for vault in BALANCER_V2_VAULTS.get(chain.id, [])]

class BalancerV2:
    def __init__(self) -> None:
        self.vaults = BALANCER_V2_VAULTS

    def is_pool(self, token_address):
        try: return BalancerV2Pool(token_address).is_pool()
        except AttributeError: return False
    
    def get_pool_price(self, pool_address, block=None):
        return BalancerV2Pool(pool_address).get_pool_price()

    def get_token_price(self, token_address, block=None):
        deepest_pool = self.deepest_pool_for(token_address, block=block)
        if deepest_pool is None: return
        return deepest_pool.get_token_price(token_address, block)
    
    def deepest_pool_for(self, token_address, block=None):
        deepest_pools = {vault.address: vault.deepest_pool_for(token_address, block=block) for vault in self.vaults}
        deepest_pools = {vault: deepest_pool for vault,deepest_pool in deepest_pools.items() if deepest_pool is not None}
        deepest_pool_balance = max(pool_balance for pool_address, pool_balance in deepest_pools.values())
        for pool_address, pool_balance in deepest_pools.values():
            if pool_balance == deepest_pool_balance and pool_address: return BalancerV2Pool(pool_address)

balancer = BalancerV2()




from ypricemagic import magic
from ypricemagic.exceptions import PriceError
from ypricemagic.classes import ERC20
from ypricemagic.utils.multicall import fetch_multicall, multicall_decimals
from ypricemagic.utils.raw_calls import _totalSupplyReadable


class BalancerV1Pool(ERC20):
    def __init__(self, pool_address) -> None:
        super().__init__(pool_address)
    
    def is_pool(self):
        required = {"getCurrentTokens", "getBalance", "totalSupply"}
        return set(self.contract.__dict__) & required == required

    def tokens(self, block=None):
        return self.contract.getCurrentTokens(block_identifier=block)

    def get_pool_price(self, block=None):
        supply = _totalSupplyReadable(self.contract, block)
        if supply == 0: return 0
        return self.get_tvl(block=block) / supply

    def get_tvl(self, block=None):
        balances = self.get_balances(block)
        return sum([balance * _get_price(token, block=block) for token, balance in balances.items()])

    def get_balances(self, block=None):
        tokens = self.tokens(block=block)
        balances = fetch_multicall(*[[self.contract, "getBalance", token] for token in tokens], block=block)
        for position, balance in enumerate(balances):
            if balance is None: balances[position] = 0
        decimals = multicall_decimals(tokens, block)
        return {token:balance / 10 ** decimal for token, balance, decimal in zip(tokens,balances,decimals)}

def _get_price(token_address, block=None):
    try: return magic.get_price(token_address, block)
    except PriceError: return 0




from ypricemagic.utils.cache import memory
from ypricemagic.utils.contracts import Contract

POOLS = {
    "0x86aFa7ff694Ab8C985b79733745662760e454169": Contract("0xF16D312d119c13dD27fD0dC814b0bCdcaAa62dfD"), # Belt.fi bDAI/bUSDC/bUSDT/bBUSD
    "0x9cb73F20164e399958261c289Eb5F9846f4D1404": Contract("0xAEA4f7dcd172997947809CE6F12018a6D5c1E8b6"), # 4Belt
}

@memory.cache()
def is_belt_lp(token_address):
    return token_address in POOLS

def get_price(token_address, block=None):
    pool = POOLS[token_address]
    return pool.get_virtual_price(block_identifier=block) / 10 ** 18




from ypricemagic import magic
from ypricemagic.utils.cache import memory
from ypricemagic.utils.contracts import Contract
from ypricemagic.utils.multicall import fetch_multicall


@memory.cache()
def is_mstable_feeder_pool(address: str) -> bool:
    contract = Contract(address)
    return hasattr(contract,'getPrice') and hasattr(contract,'mAsset')

def get_price(address, block=None):
    contract = Contract(address)
    ratio, masset, decimals = fetch_multicall([contract,'getPrice'],[contract,'mAsset'],[contract,'decimals'],block=block)
    ratio = ratio[0] / 10 ** decimals
    underlying_price = magic.get_price(masset,block)
    return underlying_price * ratio



import logging
from typing import Union

from brownie import chain, convert, multicall
from eth_typing.evm import Address, BlockNumber

from ypricemagic import _symbol
from ypricemagic.constants import STABLECOINS, WRAPPED_GAS_COIN
from ypricemagic.exceptions import PriceError
from ypricemagic.price_modules import (aave, compound, froyo,
                                       mstablefeederpool, uniswap, yearn)
from ypricemagic.price_modules.balancer.balancer import balancer
from ypricemagic.price_modules.chainlink.chainlink import chainlink
from ypricemagic.price_modules.curve import curve
from ypricemagic.utils.cache import memory
from ypricemagic.utils.contracts import Contract

logger = logging.getLogger(__name__)



@memory.cache()
def get_price(token: Union[str,Address,Contract], block: Union[BlockNumber,int,None]=None, silent: bool=False) -> float:
    token = convert.to_address(str(token))
    try: return _get_price(token, block=block,silent=silent)
    except RecursionError: raise PriceError(f'could not fetch price for {_symbol(token)} {token}')

    
def _get_price(token: Union[str,Address,Contract], block: Union[BlockNumber,int,None]=None,silent: bool=False):
    logger.debug(f"[chain{chain.id}] token: {token}")
    logger.debug("unwrapping %s", token)

    price = None

    if token == "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE":
        token = WRAPPED_GAS_COIN

    if token in STABLECOINS:
        logger.debug("stablecoin -> %s", 1)
        return 1
    
    if balancer.is_balancer_pool(token): price = balancer.get_price(token, block)
    elif token in chainlink: price = chainlink.get_price(token, block)
    elif curve and token in curve: price = curve.get_price(token, block)
    
    # these return type(price) == list
    elif yearn.is_yearn_vault(token): price = yearn.get_price(token, block=block)
        
    
    # if type(price) == list, this will output final price
    if isinstance(price, list):
        price, underlying = price
        logger.debug("peel %s %s", price, underlying)
        return price * get_price(underlying, block=block)


    if chain.id == 1: # eth mainnet
        from ypricemagic.price_modules import (cream, gelato, mooniswap,
                                               piedao, tokensets, wsteth)
        multicall(address='0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696')

        

        # we can exit early with known tokens
        if wsteth.is_wsteth(token):
            price = wsteth.get_price(token, block=block)
            logger.debug("wsteth -> %s", price)

        elif aave.is_atoken(token):
            price = aave.get_price(token, block=block)
            logger.debug("atoken -> %s", price)

        elif cream.is_creth(token):
            price = cream.get_price_creth(token, block)
            logger.debug("atoken -> %s", price)

        elif compound.is_compound_market(token):
            price = compound.get_price(token, block=block)
            logger.debug("compound -> %s", price)

        elif uniswap.is_uniswap_pool(token):
            price = uniswap.lp_price(token, block=block)
            logger.debug("uniswap pool -> %s", price)

        elif mooniswap.is_mooniswap_pool(token):
            price = mooniswap.get_pool_price(token, block=block)
            logger.debug("mooniswap pool -> %s", price)

        elif mstablefeederpool.is_mstable_feeder_pool(token):
            price = mstablefeederpool.get_price(token,block=block)
            logger.debug("mstable feeder pool -> %s", price)

        elif tokensets.is_token_set(token):
            price = tokensets.get_price(token, block=block)
            logger.debug("token set -> %s", price)

        elif piedao.is_pie(token):
            price = piedao.get_price(token, block=block)
            logger.debug("piedeo -> %s", price)

        elif gelato.is_gelato_pool(token):
            price = gelato.get_price(token, block=block)
            logger.debug("gelato -> %s", price)
        
        # peel a layer from [multiplier, underlying]
        if isinstance(price, list):
            price, underlying = price
            logger.debug("peel %s %s", price, underlying)
            return price * get_price(underlying, block=block)

        # a few more attempts to fetch a price a token
        if price is None: # NOTE: 'if not price' returns True if price == 0 but we actually only want to proceed if price == None
            price = uniswap.get_price(token, router="sushiswap", block=block)
            logger.debug("sushiswap -> %s", price)
        
        if price is None:
            price = uniswap.get_price(token, router="uniswap", block=block)
            logger.debug("uniswap -> %s", price)
        
        if price is None:
            price = uniswap.get_price_v1(token, block=block)
            logger.debug("uniswap v1 -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="shibaswap", block=block)

    if chain.id == 56: # binance smart chain
        from ypricemagic.price_modules import belt, ellipsis, ib, mooniswap

        # we can exit early with known tokens
        if belt.is_belt_lp(token):
            price = belt.get_price(token,block=block)
            logger.debug("belt -> %s", price)

        elif compound.is_compound_market(token):
            price = compound.get_price(token, block=block)
            logger.debug("compound -> %s", price)

        elif uniswap.is_uniswap_pool(token):
            price = uniswap.lp_price(token, block=block)
            logger.debug("uniswap pool -> %s", price)

        elif ib.is_ib_token(token):
            price = ib.get_price(token,block=block)
            logger.debug("ib token -> %s", price)

        elif ellipsis.is_eps_rewards_pool(token):
            price = ellipsis.get_price(token, block=block)
            logger.debug("ellipsis pool -> %s", price)

        elif mooniswap.is_mooniswap_pool(token):
            price = mooniswap.get_pool_price(token, block=block)
            logger.debug("mooniswap pool -> %s", price)

        # peel a layer from [multiplier, underlying]
        if isinstance(price, list):
            price, underlying = price
            logger.debug("peel %s %s", price, underlying)
            return price * get_price(underlying, block=block)
        
        if price is None:
            price = uniswap.get_price(token, router="pancakeswapv2", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="pancakeswapv1", block=block)
            logger.debug("uniswap -> %s", price)
        
        if price is None:
            price = uniswap.get_price(token, router="apeswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="wault", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="swapliquidity", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="thugswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="mdex", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="bakeryswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="nyanswop", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="narwhalswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="cafeswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="jetswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="babyswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="annex", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="viralata", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="elk", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="pantherswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="complus", block=block)
            logger.debug("uniswap -> %s", price)

    if chain.id == 137: # polygon
        # we can exit early with known tokens

        if aave.is_atoken(token):
            price = aave.get_price(token, block=block)
            logger.debug("atoken -> %s", price)

        elif compound.is_compound_market(token):
            price = compound.get_price(token, block=block)
            logger.debug("compound -> %s", price)

        elif mstablefeederpool.is_mstable_feeder_pool(token):
            price = mstablefeederpool.get_price(token,block=block)
            logger.debug("mstable feeder pool -> %s", price)

        elif uniswap.is_uniswap_pool(token):
            price = uniswap.lp_price(token, block=block)
            logger.debug("uniswap pool -> %s", price)

        # peel a layer from [multiplier, underlying]
        if isinstance(price, list):
            price, underlying = price
            logger.debug("peel %s %s", price, underlying)
            return price * get_price(underlying, block=block)
        
        if price is None:
            price = uniswap.get_price(token, router="quickswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="sushi", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="apeswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="dfyn", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="wault", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="cometh", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="jetswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="polyzap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="cafeswap", block=block)
            logger.debug("uniswap -> %s", price)

    if chain.id == 250: # fantom
        
        # we can exit early with known tokens

        if uniswap.is_uniswap_pool(token):
            price = uniswap.lp_price(token, block)
            logger.debug("uniswap -> %s", price)

        elif froyo.is_froyo(token):
            price = froyo.get_price(token, block=block)
            logger.debug("froyo -> %s", price)

        # peel a layer from [multiplier, underlying]
        if isinstance(price, list):
            price, underlying = price
            logger.debug("peel %s %s", price, underlying)
            return price * get_price(underlying, block=block)
        
        if price is None:
            price = uniswap.get_price(token, router="sushi", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="spookyswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="spiritswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="paintswap", block=block)
            logger.debug("uniswap -> %s", price)

        if price is None:
            price = uniswap.get_price(token, router="jetswap", block=block)
            logger.debug("uniswap -> %s", price)

    # if type(price) == list, this will output final price
    if isinstance(price, list):
        price, underlying = price
        logger.debug("peel %s %s", price, underlying)
        return price * get_price(underlying, block=block)

    # let's try a few more things
    if price is None:
        price = balancer.get_price(token, block=block)
        logger.debug("balancer -> %s", price)

    if price is None and silent is False:
        logger.error("failed to get price for %s", token)

    if price is None:
        raise PriceError(f'could not fetch price for {token}')

    return price